#include <stdlib.h>
#include <signal.h>
#include <pthread.h>
#include <unistd.h>
#include <fcntl.h>
#include <pigpio.h>
#include <verilated.h>

#include <sys/socket.h>
#include <netinet/in.h> 
#include <netinet/tcp.h>
#include <netdb.h>


//#include "Vthing.h"
#include "Vpanel.h"
//#include "Vtx0.h"

int LED_ROWS[] = {2, 3, 4};
int SWITCH_ROWS[] = {17, 27};
int COLUMNS[] = {26, 19, 13, 6, 5, 11,
	14, 15, 18, 23, 24, 25, 8, 7, 12, 16, 20, 21};

#define nelem(array) (sizeof(array)/sizeof(array[0]))

void
disablePin(int pin)
{
	gpioSetMode(pin, PI_INPUT);
	gpioSetPullUpDown(pin, PI_PUD_OFF);
}

void
enablePin(int pin)
{
	gpioSetMode(pin, PI_INPUT);
	gpioSetPullUpDown(pin, PI_PUD_UP);
}

void
outputPin(int pin, int val)
{
	gpioSetMode(pin, PI_OUTPUT);
	gpioWrite(pin, val);
}

void
delay(int ms)
{
	usleep(ms * 1000);
}



void
disableRows(void)
{
	int i;
	for(i = 0; i < nelem(LED_ROWS); i++)
		disablePin(LED_ROWS[i]);
	for(i = 0; i < nelem(SWITCH_ROWS); i++)
		disablePin(SWITCH_ROWS[i]);
}

void
enableRows(void)
{
	int i;
	for(i = 0; i < nelem(LED_ROWS); i++)
		outputPin(LED_ROWS[i], 1);
	for(i = 0; i < nelem(SWITCH_ROWS); i++)
		outputPin(SWITCH_ROWS[i], 1);
}

void
disableColumns(void)
{
	int i;
	for(i = 0; i < nelem(COLUMNS); i++)
		disablePin(COLUMNS[i]);
}

void
enableColumns(void)
{
	int i;
	for(i = 0; i < nelem(COLUMNS); i++)
		enablePin(COLUMNS[i]);
}



typedef struct Panel Panel;
struct Panel
{
	int sw0;
	int sw1;
	int lights0;
	int lights1;
	int lights2;
};

void
setLights(Panel *p)
{
	int i;

	outputPin(LED_ROWS[0], 1);
	for(i = 0; i < 18; i++)
		outputPin(COLUMNS[i], (~p->lights0>>i)&1);
	delay(2);
	outputPin(LED_ROWS[0], 0);

	outputPin(LED_ROWS[1], 1);
	for(i = 0; i < 18; i++)
		outputPin(COLUMNS[i], (~p->lights1>>i)&1);
	delay(2);
	outputPin(LED_ROWS[1], 0);

	outputPin(LED_ROWS[2], 1);
	for(i = 0; i < 18; i++)
		outputPin(COLUMNS[i], (~p->lights2>>i)&1);
	delay(2);
	outputPin(LED_ROWS[2], 0);
}

void
readSwitches(Panel *p)
{
	int i, sw;

	enableColumns();

	outputPin(SWITCH_ROWS[0], 0);
	delay(2);
	sw = 0;
	for(i = 0; i < 18; i++)
		sw |= !gpioRead(COLUMNS[i]) << i;
	disablePin(SWITCH_ROWS[0]);
	p->sw0 = sw;

	outputPin(SWITCH_ROWS[1], 0);
	delay(2);
	sw = 0;
	for(i = 0; i < 18; i++)
		sw |= !gpioRead(COLUMNS[i]) << i;
	disablePin(SWITCH_ROWS[1]);
	p->sw1 = sw;
}

int doexit;

void*
panelthread(void *arg)
{
	Panel *p = (Panel*)arg;
	while(!doexit) {
		setLights(p);
		readSwitches(p);
	}

	gpioTerminate();
	exit(0);
}

void
interrupt(int sig)
{
	doexit = 1;
}

void
panic(const char *fmt, ...)
{
        va_list ap;
        va_start(ap, fmt);
        vfprintf(stderr, fmt, ap);
        fprintf(stderr, "\n");
        va_end(ap);
        exit(1);
}
int
hasinput(int fd)
{
	fd_set fds;
	struct timeval timeout;

	if(fd < 0) return 0;

	timeout.tv_sec = 0;
	timeout.tv_usec = 0;
	FD_ZERO(&fds);
	FD_SET(fd, &fds);
	return select(fd+1, &fds, NULL, NULL, &timeout) > 0;
}
int
dial(const char *host, int port)
{
	char portstr[32];
	int sockfd;
	struct addrinfo *result, *rp, hints;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;

	snprintf(portstr, 32, "%d", port);
	if(getaddrinfo(host, portstr, &hints, &result)){
		perror("error: getaddrinfo");
		return -1;
	}

	for(rp = result; rp; rp = rp->ai_next){
		sockfd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
		if(sockfd < 0)
			continue;
		if(connect(sockfd, rp->ai_addr, rp->ai_addrlen) >= 0)
			goto win;
		close(sockfd);
	}
	freeaddrinfo(result);
	perror("error");
	return -1;

win:
	freeaddrinfo(result);
	return sockfd;
}
void
nodelay(int fd)
{
	int flag = 1;
	setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &flag, sizeof(flag));
}

struct Flexowriter
{
	int in, out;
	int dly;	// hack
};
void
putfl(Flexowriter *fl, int c) 
{
	char cc;
	cc = c & 077;
	write(fl->out, &cc, 1);
}
int
getfl(Flexowriter *fl)  
{
	char c;
	read(fl->in, &c, 1); 
	return c;
}


void
readmem(FILE *f, Vpanel *dev)
{
	char buf[100], *s;
	int a, w, par;
	int i;

	a = 0;
	while(s = fgets(buf, 100, f)) {
		while(*s) {
			if(*s == ';')
				break;
			else if('0' <= *s && *s <= '7') {
				w = strtol(s, &s, 8);
				if(*s == ':') {
					a = w;
					s++;
				} else if(a < 0200000) {
					par = ~0;
					for(i = 0; i < 18; i++)
						par ^= w<<(i+1);
					dev->panel__DOT__tx0__DOT__core[a++] = (w & 0777777) | (par&01000000);
//printf("%o: %o\n", a-1, dev->panel__DOT__tx0__DOT__core[a-1]);
				}
			}else
				s++;
		}
	}
}

int
main()
{
	Panel panel;
	Flexowriter fl;
	pthread_t th;

        if(gpioInitialise() < 0)
                return 1;
	signal(SIGINT, interrupt);
	disableRows();
	disableColumns();

	memset(&panel, 0, sizeof(panel));


	memset(&fl, 0, sizeof(fl));
	fl.in = open("/tmp/fl", O_RDWR);
	if(fl.in < 0)
		printf("can't open /tmp/fl\n");
	fl.out = fl.in;

	int disp = dial("169.254.143.255", 3400);
//	int disp = dial("dial", 3400);
	if(disp < 0)
		printf("can't open display\n");
	nodelay(disp);


	Vpanel *dev = new Vpanel;

	FILE *f;
	if(f = fopen("ut3.mem", "r")) {
		readmem(f, dev);
		fclose(f);
	}
	if(f = fopen("circle.mem", "r")) {
//	if(f = fopen("out.mem", "r")) {
		readmem(f, dev);
		fclose(f);
	}


	pthread_create(&th, NULL, panelthread, &panel);

	int fl_to_lr = 0;
	int fl_complete = 0;
	int display_complete = 0;
	int disbuf[64];
	int ndis = 0;

//panel__DOT__tx0__DOT__core

	dev->clk = 0;
	dev->reset = 1;
	for(int i = 0; i < 20; i++) {
		dev->clk ^= 1;
		dev->eval();
	}
	dev->panel__DOT__ac = rand() & 0777777;
	dev->panel__DOT__lr = rand() & 0777777;
	dev->panel__DOT__mbr = rand() & 0777777;
	dev->panel__DOT__mar = rand() & 0177777;
	dev->panel__DOT__pc = rand() & 0177777;
	dev->panel__DOT__ir = rand() & 3;
	dev->flexo_complete = 0;
	dev->flexo_to_lr = 0;

	for(;;) {
		dev->clk ^= 1;
		dev->sw0 = panel.sw0;
		dev->sw1 = panel.sw1;
		if(dev->sw_power) {
			panel.lights0 = dev->light0;
			panel.lights1 = dev->light1;
			panel.lights2 = dev->light2;
			dev->reset = 0;
		} else {
			dev->panel__DOT__ac = rand() & 0777777;
			dev->panel__DOT__lr = rand() & 0777777;
			dev->panel__DOT__mbr = rand() & 0777777;
			dev->panel__DOT__mar = rand() & 0177777;
			dev->panel__DOT__pc = rand() & 0177777;
			dev->panel__DOT__ir = rand() & 3;
			dev->reset = 1;
			panel.lights0 = 0;
			panel.lights1 = 0;
			panel.lights2 = 0;
		}
		dev->eval();

		/* flexo read */
		if(fl.dly == 0) {
			if(hasinput(fl.in)) {
				int c = getfl(&fl);
				fl_to_lr = 20;
				dev->flexo_in = c&077;
			}
			fl.dly = 10000;
		} else
			fl.dly--;
		if(fl_to_lr) fl_to_lr--;
		dev->flexo_to_lr = fl_to_lr != 0;

		/* flexo print */
		if(dev->flexo_start_print)
			fl_complete = 100;
		if(fl_complete == 50)
			putfl(&fl, dev->flexo_out);
		if(fl_complete) fl_complete--;
		dev->flexo_complete = fl_complete != 0 && fl_complete < 50;

		/* display */
		if(dev->display_start) {
//printf("%o %o\n", dev->display_x, dev->display_y);
			if(disp >= 0) {
				int x = dev->display_x;
				int y = dev->display_y;
				if(x & 0400) x++;
				if(y & 0400) y++;
				x = ((x+0400)&0777)<<1;
				y = ((y+0400)&0777)<<1;
				int cmd = x | (y<<10) | (7<<20);
disbuf[ndis++] = cmd;
if(ndis == nelem(disbuf)) {
write(disp, disbuf, 4*nelem(disbuf));
ndis = 0;
}
//				write(disp, &cmd, 4);
			}
			display_complete = 10;
		}
		if(display_complete) display_complete--;
		dev->display_complete = display_complete != 0;

	}

	delete dev;
	return 0;
}
